"""A FuriosaAI qunatizer."""

from typing import List, Mapping, Sequence, Tuple, Union

import furiosa_quantizer_impl
import onnx

import furiosa.common.utils
from furiosa.quantizer.calibrator import CalibrationMethod, Calibrator
from furiosa.quantizer.editor import ModelEditor, TensorType, get_output_names, get_pure_input_names

__version__ = furiosa.common.utils.get_sdk_version(__name__)

__full_version__ = f"Furiosa SDK Quantizer {__version__} (furiosa_quantizer_impl {furiosa_quantizer_impl.__version__} {furiosa_quantizer_impl.__git_short_hash__} {furiosa_quantizer_impl.__build_timestamp__})"  # pylint: disable=no-member # noqa: E501

__all__ = [
    "CalibrationMethod",
    "Calibrator",
    "get_pure_input_names",
    "get_output_names",
    "ModelEditor",
    "TensorType",
    "quantize",
]


def quantize(
    model: Union[onnx.ModelProto, bytes],  # pylint: disable=no-member
    tensor_name_to_range: Mapping[str, Union[Sequence[float], Tuple[int, List[Sequence[float]]]]],
) -> bytes:
    """Quantize an ONNX model on the basis of the range of its tensors.

    Args:
        model (onnx.ModelProto or bytes): An ONNX model to quantize.
        tensor_name_to_range (Mapping[str,
            Union[
                Sequence[float],
                Tuple[int, List[Sequence[float]]]
                ]
            ]):
            A mapping from a tensor name to calibration range.
            calibration range type is union type which has two types below.
            #. per tensor (Sequence[float])
                #. it has min-max range for tensor.
            #. per axis (Tuple[int, List[Sequence[float]]])
                #. it contains axis and list of min-max range
                #. length of the list must match the size of
                    the corresponding axis of the tensor.

    Returns:
        bytes: A serialized ONNX model that incorporates quantization
            information.
    """
    if isinstance(model, onnx.ModelProto):  # pylint: disable=no-member
        model = model.SerializeToString()
    return furiosa_quantizer_impl.quantize(model, tensor_name_to_range)  # pylint: disable=no-member



import enum
from typing import List, Optional, Tuple

import furiosa_quantizer_impl
import onnx


class TensorType(enum.IntEnum):
    """An enumeration class representing the element type of a tensor.

    This class is used with the ModelEditor.convert_{input,output}_type method to specify the
    desired element type.
    """

    UINT8 = 1
    INT8 = 2


def get_pure_input_names(model: onnx.ModelProto) -> List[str]:
    """Return the names of inputs in an ONNX model that have no associated initializers.

    Args:
        model (onnx.ModelProto): An ONNX model.

    Returns:
        List[str]: A list of the names of inputs in the model that have no associated initializers.
    """
    initializers = {tensor.name for tensor in model.graph.initializer}
    return [
        value_info.name for value_info in model.graph.input if value_info.name not in initializers
    ]


def get_output_names(model: onnx.ModelProto) -> List[str]:
    """Return the names of outputs in an ONNX model.

    Args:
        model (onnx.ModelProto): An ONNX model.

    Returns:
        List[str]: A list of the names of outputs in the model.
    """
    return [value_info.name for value_info in model.graph.output]


class ModelEditor:
    """A utility class for manipulating ONNX models."""

    def __init__(self, model: onnx.ModelProto):
        self._input = set(get_pure_input_names(model))
        self._output = set(get_output_names(model))
        self._processor = furiosa_quantizer_impl.Processor(model)

    def convert_input_type(self, tensor_name: str, tensor_type: TensorType) -> None:
        """Convert the element type of an input tensor named tensor_name to tensor_type.

        Args:
            tensor_name (str): The name of an input tensor to convert.
            tensor_type (TensorType): The desired element type.
        """
        if tensor_name not in self._input:
            raise ValueError(f"Could find a pure input tensor named '{tensor_name}'.")

        if tensor_type == TensorType.UINT8:
            self._processor.use_u8_for_input(tensor_name)
        else:
            raise ValueError(f"ModelEditor.convert_input_type does not support {tensor_type} yet.")

    def convert_output_type(
        self,
        tensor_name: str,
        tensor_type: TensorType,
        tensor_range: Optional[Tuple[float, float]] = None,
    ) -> None:
        """Convert the element type of an output tensor named tensor_name to tensor_type.

        Args:
            tensor_name (str): The name of an output tensor to convert.
            tensor_type (TensorType): The desired element type.
            tensor_range (Optional[Tuple[float, float]]): A new min/max range of the output tensor.
                If it is None, the original range will be retained. Defaults to None.
        """
        if tensor_name not in self._output:
            raise ValueError(f"Could not find an output tensor named '{tensor_name}'.")

        if tensor_type == TensorType.UINT8:
            self._processor.use_u8_for_output(tensor_name, tensor_range)
        elif tensor_type == TensorType.INT8:
            self._processor.use_i8_for_output(tensor_name, tensor_range)
        else:
            raise ValueError(f"ModelEditor.convert_output_type does not support {tensor_type} yet.")
